#!/usr/bin/env python3
"""
TCP Proxy
Transparent TCP proxy with traffic inspection capabilities.
Parameters:
    - local_host: Local host to bind
    - local_port: Local port to listen on
    - remote_host: Remote host to forward to
    - remote_port: Remote port to forward to
Requires: admin/root privileges for privileged ports
"""
import socket
import threading
import argparse
import logging
from datetime import datetime
import json
import queue
import select
import time
import hexdump

class TCPProxy:
    def __init__(self, local_host, local_port, remote_host, remote_port):
        self.local_host = local_host
        self.local_port = local_port
        self.remote_host = remote_host
        self.remote_port = remote_port
        self.connections = []
        self.traffic_log = queue.Queue()
        self.setup_logging()

    def setup_logging(self):
        logging.basicConfig(
            level=logging.INFO,
            format='%(asctime)s - %(levelname)s - %(message)s'
        )
        self.logger = logging.getLogger(__name__)

    def hexdump(self, data):
        return hexdump.hexdump(data, result='return')

    def receive_from(self, connection):
        buffer = b""
        connection.settimeout(2)
        
        try:
            while True:
                data = connection.recv(4096)
                if not data:
                    break
                buffer += data
        except:
            pass
            
        return buffer

    def analyze_request(self, data):
        """Analyze request data for known protocols"""
        try:
            # Check if it's HTTP
            if data.startswith(b'GET') or data.startswith(b'POST') or \
               data.startswith(b'PUT') or data.startswith(b'DELETE'):
                return 'HTTP Request', data.decode('utf-8', errors='ignore')
            
            # Check if it's SMTP
            if data.startswith(b'HELO') or data.startswith(b'EHLO'):
                return 'SMTP Request', data.decode('utf-8', errors='ignore')
            
            # Check if it's FTP
            if data.startswith(b'USER') or data.startswith(b'PASS'):
                return 'FTP Request', data.decode('utf-8', errors='ignore')
            
            # If no known protocol is detected, return raw hex dump
            return 'Unknown Protocol', self.hexdump(data)
            
        except Exception as e:
            self.logger.error(f"Error analyzing request: {str(e)}")
            return 'Error', str(e)

    def analyze_response(self, data):
        """Analyze response data for known protocols"""
        try:
            # Check if it's HTTP
            if data.startswith(b'HTTP/'):
                return 'HTTP Response', data.decode('utf-8', errors='ignore')
            
            # Check if it's SMTP
            if data.startswith(b'220') or data.startswith(b'250'):
                return 'SMTP Response', data.decode('utf-8', errors='ignore')
            
            # Check if it's FTP
            if data.startswith(b'230') or data.startswith(b'530'):
                return 'FTP Response', data.decode('utf-8', errors='ignore')
            
            # If no known protocol is detected, return raw hex dump
            return 'Unknown Protocol', self.hexdump(data)
            
        except Exception as e:
            self.logger.error(f"Error analyzing response: {str(e)}")
            return 'Error', str(e)

    def proxy_handler(self, client_socket, client_address):
        remote_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        
        try:
            remote_socket.connect((self.remote_host, self.remote_port))
            
            while True:
                readable, _, _ = select.select(
                    [client_socket, remote_socket], [], []
                )
                
                if client_socket in readable:
                    request = self.receive_from(client_socket)
                    if len(request) == 0:
                        break
                    
                    protocol, analyzed = self.analyze_request(request)
                    self.traffic_log.put({
                        'timestamp': datetime.now().isoformat(),
                        'source': f"{client_address[0]}:{client_address[1]}",
                        'destination': f"{self.remote_host}:{self.remote_port}",
                        'direction': 'request',
                        'protocol': protocol,
                        'size': len(request),
                        'data': analyzed
                    })
                    
                    remote_socket.send(request)
                
                if remote_socket in readable:
                    response = self.receive_from(remote_socket)
                    if len(response) == 0:
                        break
                    
                    protocol, analyzed = self.analyze_response(response)
                    self.traffic_log.put({
                        'timestamp': datetime.now().isoformat(),
                        'source': f"{self.remote_host}:{self.remote_port}",
                        'destination': f"{client_address[0]}:{client_address[1]}",
                        'direction': 'response',
                        'protocol': protocol,
                        'size': len(response),
                        'data': analyzed
                    })
                    
                    client_socket.send(response)
                    
        except Exception as e:
            self.logger.error(f"Proxy handler error: {str(e)}")
        finally:
            client_socket.close()
            remote_socket.close()

    def save_traffic_log(self, output_file):
        traffic = []
        while not self.traffic_log.empty():
            traffic.append(self.traffic_log.get())

        log_data = {
            'proxy_config': {
                'local_host': self.local_host,
                'local_port': self.local_port,
                'remote_host': self.remote_host,
                'remote_port': self.remote_port
            },
            'traffic': traffic
        }

        try:
            with open(output_file, 'w') as f:
                json.dump(log_data, f, indent=4)
            self.logger.info(f"Traffic log saved to {output_file}")
        except Exception as e:
            self.logger.error(f"Error saving traffic log: {str(e)}")

    def start_proxy(self, output_file=None):
        server = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        server.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        
        try:
            server.bind((self.local_host, self.local_port))
            server.listen(5)
            
            self.logger.info(
                f"Proxy listening on {self.local_host}:{self.local_port}"
            )
            self.logger.info(
                f"Forwarding to {self.remote_host}:{self.remote_port}"
            )
            
            while True:
                client_socket, addr = server.accept()
                self.logger.info(f"Received connection from {addr[0]}:{addr[1]}")
                
                proxy_thread = threading.Thread(
                    target=self.proxy_handler,
                    args=(client_socket, addr)
                )
                proxy_thread.daemon = True
                proxy_thread.start()
                self.connections.append(proxy_thread)
                
        except KeyboardInterrupt:
            self.logger.info("Stopping proxy...")
            if output_file:
                self.save_traffic_log(output_file)
        except Exception as e:
            self.logger.error(f"Proxy error: {str(e)}")
        finally:
            server.close()

if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="TCP Proxy")
    parser.add_argument("--local-host", default="0.0.0.0",
                      help="Local host to bind")
    parser.add_argument("--local-port", type=int, required=True,
                      help="Local port to listen on")
    parser.add_argument("--remote-host", required=True,
                      help="Remote host to forward to")
    parser.add_argument("--remote-port", type=int, required=True,
                      help="Remote port to forward to")
    parser.add_argument("--output", help="Output file for traffic log")
    args = parser.parse_args()

    proxy = TCPProxy(args.local_host, args.local_port,
                    args.remote_host, args.remote_port)
    proxy.start_proxy(args.output)
